//! This module helps to deal with controls.

use crate::engine::input::{Action, Button, Event, InputManager, Key};
use crate::engine::vector::Vector;

/// The different types of controls.
#[derive(Clone)]
pub enum Controls {
    /// Default keyboard controls.
    Keyboard,

    /// A gamepad controls the scene.
    Gamepad(u32),
}

impl Controls {
    /// Returns the main direction held by the user.
    pub fn main_direction(&self, input: &InputManager) -> Vector {
        match self {
            Controls::Keyboard => {
                let mut dir = Vector::new(0.0, 0.0);
                if input.is_key_pressed(Key::ArrowLeft) {
                    dir.x -= 1.0;
                }

                if input.is_key_pressed(Key::ArrowRight) {
                    dir.x += 1.0;
                }

                dir
            }
            Controls::Gamepad(id) => {
                if let Some(gamepad) = input.gamepad(*id) {
                    Vector::new(gamepad.main_direction().x, 0.0)
                } else {
                    Vector::new(0.0, 0.0)
                }
            }
        }
    }

    /// Converts an event into an action if it corresponds to an action.
    pub fn convert(&self, event: Event) -> Option<Action> {
        match self {
            Controls::Keyboard => match event {
                Event::KeyPressed(Key::Space) => Some(Action::ButtonPressed(Button::Button1)),
                Event::KeyReleased(Key::Space) => Some(Action::ButtonReleased(Button::Button1)),
                _ => None,
            },
            Controls::Gamepad(ig) => match event {
                Event::ButtonPressed(ie, button) if *ig == ie => {
                    Some(Action::ButtonPressed(button))
                }
                Event::ButtonReleased(ie, button) if *ig == ie => {
                    Some(Action::ButtonReleased(button))
                }
                _ => None,
            },
        }
    }
}
